import ode
from math import sin, asin, pi
from pyeuclid.euclid import Vector3, Matrix4

class Wing (object):
	crit_angle = 10 / 180. * pi
	max_angle = 30 / 180. * pi

	lift_rate = 0.2
	drag_rate = 0.05

	def __init__(self, world, normal, position, rate, weight):
		self.normal = normal.normalized()
		self.rate = rate

		self.body = ode.Body(world)
		
		mass = ode.Mass()
		mass.setSphere(1, weight)
		self.body.setMass(mass)
		self.body.setPosition(position)

		self.wind = Vector3(0, 0, 0)

	def lift_coeff(self, angle):
		mod = 1.
		if angle < 0:
			mod = -1
			angle = -angle

		if angle < self.crit_angle:
			return (1. - (self.crit_angle - angle) / self.crit_angle) ** 2 * mod
		elif angle > self.max_angle:
			return 0.
		else:
			return ((self.max_angle - angle) / (self.max_angle - self.crit_angle))**2 * mod
	
	def set_wind(self, wind):
		self.wind = wind
	
	def add_wind(self, wind):
		self.wind += wind

	def think(self):
		vel = Vector3(*self.body.vectorFromWorld(Vector3(*self.body.getLinearVel()) - self.wind))
		
		drag_direction = -vel.normalized()
		lift_direction = drag_direction.cross(self.normal).cross(drag_direction).normalized()
		angle = asin(self.normal.dot(drag_direction))
		
		amount = abs(vel) ** 2 * self.rate

		drag = drag_direction * abs(sin(angle)) * self.drag_rate
		lift = lift_direction * self.lift_coeff(angle) * self.lift_rate

		self.last_force = force = (drag + lift) * amount

		self.body.addRelForce(force)


class Flyer (object):
	gravity = Vector3(0., 0., -.5)
	rudder_rate = 0.01

	def __init__(self, pos=Vector3(0, 0, 0)):
		self.world = ode.World()
		self.world.setGravity(self.gravity)
		self.alive = 1

		self.wings = [
			# Front left
			Wing(self.world,
				Vector3(0., 0., 1.),
				Vector3(1., 1., 0.) + pos,
				.135,
				.8),

			# Front right
			Wing(self.world,
				Vector3(0., 0., 1.),
				Vector3(1., -1., 0.) + pos,
				.135,
				.8),

			# Back
			Wing(self.world,
				Vector3(0., 0., 1.),
				Vector3(-1., 0., 0.) + pos,
				.275,
				0.95),

			# Vertical left
			Wing(self.world,
				Vector3(0., 1., 0.),
				Vector3(-1., -1., 0.) + pos,
				.2,
				.5),

			# Vertial right
			Wing(self.world,
				Vector3(0., 1., 0.),
				Vector3(-1., 1., 0.) + pos,
				.2,
				.5),

			# Vertical front
			Wing(self.world,
				Vector3(0., 1., 0.),
				Vector3(.8, 0., 0.) + pos,
				.18,
				.5)
		]
		
		self.joints = []

		self.fix_wings(self.wings[0], self.wings[2])
		self.fix_wings(self.wings[1], self.wings[2])
		self.fix_wings(self.wings[3], self.wings[2])
		self.fix_wings(self.wings[4], self.wings[2])
		self.fix_wings(self.wings[5], self.wings[2])

		self.main_body = self.wings[2].body
	
	def fix_wings(self, a, b):
		joint = ode.FixedJoint(self.world)
		joint.attach(a.body, b.body)
		joint.setFixed()
		self.joints.append(joint)

	def set_wind(self, wind):
		for w in self.wings:
			w.set_wind(wind)

	def add_wind(self, wind):
		for w in self.wings:
			w.add_wind(wind)

	def think(self):
		for w in self.wings:
			w.think()

		self.world.step(1/10.)

		self.pos = Vector3(*self.main_body.getPosition())

if __name__ == '__main__':
	import os, pygame
	from pygame.draw import line, circle
	from pygame.locals import *

	import data

	pygame.init()

	disp = pygame.display.set_mode((1200, 800), DOUBLEBUF)
	font = pygame.font.Font(data.get_path(os.path.join('fonts', 'DejaVuSansMono.ttf')), 10)

	clock = pygame.time.Clock()
	white = (255, 255, 255)

	def text(pos, text):
		surf = font.render(text, 1, white)
		disp.blit(surf, pos)

	vbase = Vector3(600, 0, 50)
	scale = 10

	def draw_line(a, b, color):
		ax = int(vbase.x + a.x * scale)
		ay = int(vbase.z + a.z * scale)
		bx = int(vbase.x + b.x * scale)
		by = int(vbase.z + b.z * scale)

		line(disp, color, (600 + ax, 400 - ay), (600 + bx, 400 - by))

	def draw_wing(wing, color):
		body = wing.body
		pos = Vector3(*body.getPosition())

		rot = Matrix4()
		rot.a, rot.b, rot.c, rot.e, rot.f, rot.g, rot.i, rot.j, rot.k = body.getRotation()

		a = pos + rot * Vector3(-1, 0, 0)
		b = pos + rot * Vector3(1, 0, 0)

		draw_line(a, b, color)

		f = Vector3(*wing.body.vectorToWorld(wing.last_force)) * 4
		a = pos
		b = pos + f
		draw_line(a, b, (255, 0, 0))
	
	
	flyer = Flyer(Vector3(0, 0, 10000))

	kstate = {
		K_UP: 0,
		K_LEFT: 0,
		K_RIGHT: 0
	}

	running = 1
	while running:
		for ev in pygame.event.get():
			if ev.type == KEYDOWN and ev.key == K_ESCAPE:
				running = 0

			elif ev.type in [KEYDOWN, KEYUP]:
				kstate[ev.key] = ev.type == KEYDOWN

			elif ev.type == QUIT:
				running = 0

		steer_force = 4.

		if kstate[K_LEFT]:
			flyer.wings[2].body.addRelForce((0, 0, steer_force))
		if kstate[K_RIGHT]:
			flyer.wings[2].body.addRelForce((0, 0, -steer_force))

		flyer.think()
		disp.fill(0)

		pos = flyer.wings[2].body.getPosition()
		vel = flyer.wings[2].body.getLinearVel()

		text((805, 5), "P: %.4f %.4f %.4f" % pos)
		text((805, 15), "V: %.4f %.4f %.4f" % vel)
		text((810, 25), "|V|: %.4f" % Vector3(*vel).magnitude())

		pos = Vector3(*pos)
		vel = Vector3(*vel)

		vbase = -pos * scale

		draw_wing(flyer.wings[1], (128, 255, 0))
		draw_wing(flyer.wings[2], (255, 128, 0))
		draw_wing(flyer.wings[3], (0, 128, 255))

		draw_line(pos, pos + vel, (0, 0, 255))

		pygame.display.flip()
		clock.tick(30)

