import random
from math import sin, cos, pi

import pygame
from pygame.locals import *
from OpenGL.GL import *
from OpenGL.GLU import *

import sounds
from data import get_path
from pyeuclid.euclid import Vector3
from wind import Wind
from texture import Texture

class TriggerRing (object):
	thickness = 5.

	def compile(self):
		self.list = glGenLists(1)
		glNewList(self.list, GL_COMPILE)
		glBegin(GL_QUAD_STRIP)
		angle = 0.
		while angle <= 2 * pi:
			x, y = sin(angle), cos(angle)
			glVertex3f(x, y, 0)
			glVertex3f(x, y, self.thickness)

			angle += pi / 8

		glVertex3f(0, 1., 0)
		glVertex3f(0, 1., self.thickness)
		glEnd()
		glEndList()
	
	def render(self):
		glCallList(self.list)

class Trigger (object):
	z_offset = 50
	ring = TriggerRing()
	color = (1., 1., 1., 1.)
	speed = 10.

	def __init__(self, pos, size, height):
		self.pos = pos
		self.size = size
		self.height = height
		self.time = 0.

		self.top = self.pos.z + self.height

	def render(self):
		glPushMatrix()
		glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT)
		glDisable(GL_TEXTURE_2D)

		self.time += self.speed / 10.
		self.time %= self.z_offset

		glTranslatef(self.pos.x, self.pos.y, self.pos.z + self.time)
		glScalef(self.size, self.size, 1)
		glColor4fv(self.color)

		for z in range(0, self.height, self.z_offset):
			self.ring.render()
			glTranslatef(0, 0, self.z_offset)
		
		glPopAttrib()
		glPopMatrix()
	
	def is_inside(self, pos):
		proj = Vector3(pos.x, pos.y, self.pos.z)
		return (self.pos.z <= pos.z <= self.top and (proj - self.pos).magnitude() < self.size)

	def action(self, flyer):
		pass


class Draught (Trigger):
	rate = 0.2

	def __init__(self, pos, size, height, speed):
		Trigger.__init__(self, pos, size, height)
		self.speed = speed

		if speed > 0:
			self.color = (1., 0., 0., .2)
		else:
			self.color = (0, 0, 1., .2)

		self.att_radius = self.size * (1 - self.rate)
		self.att_height1 = self.height * self.rate
		self.att_height2 = self.height * (1 - self.rate)


	def action(self, flyer):
		pos = flyer.pos
		h = pos.z - self.pos.z
		d = (Vector3(pos.x, pos.y, self.pos.z) - self.pos).magnitude()

		att = 1.
		
		if h < self.att_height1:
			att *= h / self.att_height1
		if h > self.att_height2:
			att *= 1 - (h - self.att_height2) / self.att_height1
		if d > self.att_radius:
			att *= 1 - (d - self.att_radius) / (self.size - self.att_radius)

		flyer.add_wind((0, 0, self.speed * att))

class Area (object):
	seed = 1252163114
	area_size = 10000
	area_grid = 5
	grid_size = 1000
	richness = (4, 5)
	draughts = [-5, 5, 5, 5, 10, 10]

	def __init__(self, x, y):
		self.x, self.y = x, y
		self.triggers = []
		self.generate()

	def get_area_id(self):
		return self.seed + self.x * self.area_size + self.y

	def generate(self):
		random.seed(self.get_area_id())

		self.ndraughts = random.randint(*self.richness)
		for i in range(self.ndraughts):
			x = random.randint(1, self.area_grid) * self.grid_size + self.x * self.area_size
			y = random.randint(1, self.area_grid) * self.grid_size + self.y * self.area_size
			z = random.choice([0, 800])

			speed = random.choice(self.draughts)
			draught = Draught(Vector3(x, y, z), 500, 750, speed)

			self.triggers.append(draught)
	
	def needed(self, pos):
		x = pos.x // self.area_size - self.x
		y = pos.y // self.area_size - self.y

		return ((-2 < x < 2) and (-2 < y < 2))

class Environment (object):
	color_sky = Vector3(.7, .7, 1.)
	color_clouds = Vector3(1., 1., 1.)
	color_water = Vector3(.15, .25, .28)
	color_deep = Vector3(0., 0., 0.)
	fog_start = 4000
	fog_end = 8000
	cloud_thickness = 100
	cloudline = 800
	cloudline_low = cloudline - cloud_thickness
	cloudline_high = cloudline + cloud_thickness
	waterline = 2.
	deepline = -50.

	def setup(self):
		glFogi(GL_FOG_MODE, GL_LINEAR)
		glEnable(GL_FOG)

	def draw_quad(self):
		d = Area.area_size / 2
		glTexCoord2f(0, 0)
		glVertex3f(-d, -d, 0)
		glTexCoord2f(5, 0)
		glVertex3f(d, -d, 0)
		glTexCoord2f(5, 5)
		glVertex3f(d, d, 0)
		glTexCoord2f(0, 5)
		glVertex3f(-d, d, 0)

	def compile(self):
		# Dirty hack.
		Trigger.ring.compile()

		self.textures = [Texture(), Texture()]
		
		self.textures[0].load('water.png')
		self.textures[1].load('broken.png')

		self.list = glGenLists(1)
		glNewList(self.list, GL_COMPILE)

		self.textures[0].select()
		glBegin(GL_QUADS)
		self.draw_quad()
		glEnd()
		self.textures[1].select()
		glTranslatef(0, 0, self.cloudline)
		glBegin(GL_QUADS)
		self.draw_quad()
		glEnd()
		glEndList()

	def render_view(self, pos, face, up):
		if self.cloudline_low < pos[2] < self.cloudline_high:
			d = abs(pos[2] - self.cloudline) / self.cloud_thickness
			glFogf(GL_FOG_START, d**2 * self.fog_start + 1)
			glFogf(GL_FOG_END, d * self.fog_end + 1)
		elif self.deepline < pos[2] < 0:
			d = 1 - pos[2] / self.deepline
			glFogf(GL_FOG_START, d**3 * self.fog_start + 10)
			glFogf(GL_FOG_END, d**2 * self.fog_end + 150)
		elif pos[2] < self.deepline:
			glFogf(GL_FOG_START, 10)
			glFogf(GL_FOG_END, 150)
		else:
			glFogf(GL_FOG_START, self.fog_start)
			glFogf(GL_FOG_END, self.fog_end)


		if self.cloudline < pos[2] < self.cloudline_high:
			color = (self.color_sky * d + self.color_clouds * (1 - d))
		elif self.deepline < pos[2] < 0:
			color = (self.color_water * d + self.color_deep * (1 - d))
		elif pos[2] < self.deepline:
			color = self.color_deep
		elif pos[2] < 800:
			color = self.color_clouds
		else:
			color = self.color_sky

		glFogfv(GL_FOG_COLOR, tuple(color))
		glClearColor(color.x, color.y, color.z, 1)

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
		glLoadIdentity()
		gluLookAt(pos[0], pos[1], pos[2],
			pos[0] + face[0], pos[1] + face[1], pos[2] + face[2],
			up[0], up[1], up[2])


		as = Area.area_size

		cx = pos[0] // as
		cy = pos[1] // as

		for x in range(-2, 3, 1):
			for y in range(-2, 3, 1):
				glPushMatrix()
				glTranslatef((cx + x) * as, (cy + y) * as, 0)
				glCallList(self.list)
				glPopMatrix()


class Endless (Environment):
	def __init__(self):
		self.face = Vector3(1, 0, 0)
		self.up = Vector3(0, 0, 1)
		self.pos = Vector3(0, 0, 1200)
	
	def think(self):
		self.pos.x += 5
	
	def render(self):
		self.render_view(self.pos, self.face, self.up)

class World (Environment):
	spawn_pos = Vector3(-8000, 2200, 800)

	def __init__(self, flyer):
		self.flyer = flyer
		self.wind = Wind()
		self.areas = {}
		self.triggers = []

	def setup(self):
		Environment.setup(self)
		self.wind.play()
	
	def get_current_area(self):
		x = self.flyer.pos.x // Area.area_size
		y = self.flyer.pos.y // Area.area_size
		return (x, y)
	
	def check_areas(self):
		cx, cy = self.get_current_area()
		self.triggers = []

		for (x, y), area in self.areas.items():
			if not (-2 < (x - cx) < 2 and -2 < (x - cy) < 2):
				del self.areas[(x, y)]

		for dx in range(-1, 2):
			for dy in range(-1, 2):
				x, y = cx + dx, cy + dy
				
				if (x, y) not in self.areas:
					self.areas[(x, y)] = Area(x, y)

				self.triggers.extend(self.areas[(x, y)].triggers)

	def think(self):
		self.flyer.think()
		pos = self.flyer.pos

		self.check_areas()

		self.flyer.set_wind(Vector3(0, 0, 0))
		for t in self.triggers:
			if t.is_inside(pos):
				t.action(self.flyer)

		if pos.z < self.waterline:
			if self.flyer.alive:
				self.wind.stop()
				sounds.play_sample('splash')

			self.flyer.alive = 0
			for w in self.flyer.wings:
				r = -Vector3(*w.body.getLinearVel()) * 0.5
				w.body.addForce(r)




	def render(self):
		wing = self.flyer.main_body
		pos = wing.getPosition()
		face = wing.vectorToWorld((1, 0, 0))
		up = wing.vectorToWorld((0, 0, 1))

		self.wind.set_speed(Vector3(*wing.getLinearVel()).magnitude())

		self.render_view(pos, face, up)

		for t in self.triggers:
			t.render()



if __name__ == '__main__':

	from flyer import Flyer

	pygame.mixer.pre_init(44100)
	pygame.init()

	disp = pygame.display.set_mode((1024, 768), OPENGL | DOUBLEBUF)
	glMatrixMode(GL_PROJECTION)
	glLoadIdentity()
	gluPerspective(45, 1024./768., 1, 10000)
	glMatrixMode(GL_MODELVIEW)
	glEnable(GL_TEXTURE_2D)
	glEnable(GL_DEPTH_TEST)


	clock = pygame.time.Clock()

	flyer = Flyer(Vector3(-4000, 0, 150))
	world = World(flyer)

	world.setup()
	world.compile()

	for i in range(10):
		pos = Vector3(random.uniform(-4000, 4000), random.uniform(-4000, 4000), random.randint(0, 1) * 800)
		speed = random.choice([-2, -1, 1, 2]) * 3
		draught = Draught(pos, 800, 800, speed)
		world.triggers.append(draught)

	kstate = dict.fromkeys([K_LEFT, K_RIGHT, K_UP, K_DOWN, K_SPACE, K_LSHIFT])


	running = 1
	while running:
		for ev in pygame.event.get():
			if ev.type == KEYDOWN and ev.key == K_ESCAPE:
				running = 0
			if ev.type in [KEYUP, KEYDOWN]:
				kstate[ev.key] = ev.type == KEYDOWN
			if ev.type == QUIT:
				running = 0
		
		h_force = .3
		v_force = 2.
		throttle = 10.
		
		if kstate[K_LEFT]:
			flyer.wings[0].body.addRelForce((0, 0, -h_force))
			flyer.wings[1].body.addRelForce((0, 0, h_force))
		if kstate[K_RIGHT]:
			flyer.wings[0].body.addRelForce((0, 0, h_force))
			flyer.wings[1].body.addRelForce((0, 0, -h_force))
		if kstate[K_UP]:
			flyer.wings[2].body.addRelForce((0, 0, v_force))
		if kstate[K_DOWN]:
			flyer.wings[2].body.addRelForce((0, 0, -v_force))
		if kstate[K_SPACE]:
			flyer.wings[0].body.addRelForce((throttle, 0, 0))
			flyer.wings[1].body.addRelForce((throttle, 0, 0))

		world.think()
		world.render()

		pygame.display.flip()
		clock.tick(25)
